/*
 * Javolution - Java(TM) Solution for Real-Time and Embedded Systems
 * Copyright (C) 2006 - Javolution (http://javolution.org/)
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software is
 * freely granted, provided that this notice is preserved.
 */
package javolution.text;

import j2me.lang.CharSequence;
import j2me.lang.ThreadLocal;
import javolution.text.TextFormat.Cursor;

import java.io.IOException;

/**
 * <p> This class provides utility methods to parse <code>CharSequence</code>
 *     into primitive types and to format primitive types into an 
 *     <code>Appendable</code>.</p>
 *
 * <p> Methods from this class <b>do not create temporary objects</b>
 *     and are typically faster than standard library methods (e.g
 *     <code>TypeFormat#parseDouble</code> is up to 15x faster than 
 *     <code>Double.parseDouble</code>).</p>
 *
 * <p> For non-primitive objects, formatting is typically performed using 
 *     specialized {@link TextFormat} instances.</p>
 * 
 * @author  <a href="mailto:jean-marie@dautelle.com">Jean-Marie Dautelle</a>
 * @version 3.7, January 1, 2006
 */
public final class TypeFormat {

    /**
     * Holds the reusable cursor. 
     */
    final static ThreadLocal CURSOR = new ThreadLocal() {
        protected Object initialValue() {
            return new Cursor();
        }
    };

    /**
     * Holds the digits to character mapping. 
     */
    private final static char[] DIGIT_TO_CHAR = { '0', '1', '2', '3', '4', '5',
            '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i',
            'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
            'w', 'x', 'y', 'z' };

    /**
     * Default constructor (forbids derivation).
     */
    private TypeFormat() {
    }

    /**
     * Parses the specified character sequence as a <code>boolean</code>.
     *
     * @param  csq the character sequence to parse.
     * @return the corresponding boolean value.
     * @throws IllegalArgumentException if the specified character sequence 
     *         is different from "true" or "false" ignoring cases.
     */
    public static boolean parseBoolean(CharSequence csq) {
        if ((csq.length() == 4)
                && (csq.charAt(0) == 't' || csq.charAt(0) == 'T')
                && (csq.charAt(1) == 'r' || csq.charAt(1) == 'R')
                && (csq.charAt(2) == 'u' || csq.charAt(2) == 'U')
                && (csq.charAt(3) == 'e' || csq.charAt(3) == 'E')) {
            return true;
        } else if ((csq.length() == 5)
                && (csq.charAt(0) == 'f' || csq.charAt(0) == 'F')
                && (csq.charAt(1) == 'a' || csq.charAt(1) == 'A')
                && (csq.charAt(2) == 'l' || csq.charAt(2) == 'L')
                && (csq.charAt(3) == 's' || csq.charAt(3) == 'S')
                && (csq.charAt(4) == 'e' || csq.charAt(4) == 'E')) {
            return false;
        }
        throw new IllegalArgumentException("Cannot parse " + csq
                + " as boolean");
    }

    /**
     * Parses the specified character sequence from the specified position 
     * as a <code>boolean</code>.
     *
     * @param csq the character sequence to parse.
     * @param cursor the current cursor position (being maintained).
     * @return the next boolean value.
     * @throws IllegalArgumentException if the character sequence from the 
     *         specified position is different from "true" or "false" ignoring
     *         cases.
     */
    public static boolean parseBoolean(CharSequence csq, Cursor cursor) {
        final int i = cursor.getIndex();
        if ((csq.length() > i + 4)
                && (csq.charAt(i) == 't' || csq.charAt(i) == 'T')
                && (csq.charAt(i + 1) == 'r' || csq.charAt(i + 1) == 'R')
                && (csq.charAt(i + 2) == 'u' || csq.charAt(i + 2) == 'U')
                && (csq.charAt(i + 3) == 'e' || csq.charAt(i + 3) == 'E')) {
            cursor.increment(4);
            return true;
        }
        if ((csq.length() > i + 5)
                && (csq.charAt(i) == 'f' || csq.charAt(i) == 'F')
                && (csq.charAt(i + 1) == 'a' || csq.charAt(i + 1) == 'A')
                && (csq.charAt(i + 2) == 'l' || csq.charAt(i + 2) == 'L')
                && (csq.charAt(i + 3) == 's' || csq.charAt(i + 3) == 'S')
                && (csq.charAt(i + 4) == 'e' || csq.charAt(i + 4) == 'E')) {
            cursor.increment(5);
            return false;
        }
        throw new IllegalArgumentException("Cannot parse boolean at "
                + cursor.getIndex());
    }

    /**
     * Parses the specified character sequence as a signed decimal 
     * <code>byte</code>.
     *
     * @param  csq the character sequence to parse.
     * @return <code>parseByte(csq, 10)</code>
     * @throws NumberFormatException if the specified character sequence
     *         does not contain a parsable <code>byte</code>.
     * @see    #parseByte(CharSequence, int)
     */
    public static byte parseByte(CharSequence csq) {
        return parseByte(csq, 10);
    }

    /**
     * Parses the specified character sequence as a signed <code>byte</code> 
     * in the specified radix.
     *
     * @param  csq the character sequence to parse.
     * @param  radix the radix to be used while parsing.
     * @return the corresponding <code>byte</code>.
     * @throws NumberFormatException if the specified character sequence
     *         does not contain a parsable <code>byte</code>.
     */
    public static byte parseByte(CharSequence csq, int radix) {
        int i = parseInt(csq, radix);
        if ((i < Byte.MIN_VALUE) || (i > Byte.MAX_VALUE))
            throw new NumberFormatException("Overflow");
        return (byte) i;
    }

    /**
     * Parses the specified character sequence from the specified position 
     * as a signed <code>byte</code> in the specified radix.
     *
     * @param  csq the character sequence to parse.
     * @param  radix the radix to be used while parsing.
     * @param cursor the current cursor position (being maintained).
     * @return the corresponding <code>byte</code>.
     * @throws NumberFormatException if the specified character sequence
     *         does not contain a parsable <code>byte</code>.
     */
    public static byte parseByte(CharSequence csq, int radix, Cursor cursor) {
        int i = parseInt(csq, radix, cursor);
        if ((i < Byte.MIN_VALUE) || (i > Byte.MAX_VALUE))
            throw new NumberFormatException("Overflow");
        return (byte) i;
    }

    /**
     * Parses the specified character sequence as a signed decimal 
     * <code>short</code>.
     *
     * @param  csq the character sequence to parse.
     * @return <code>parseShort(csq, 10)</code>
     * @throws NumberFormatException if the specified character sequence
     *         does not contain a parsable <code>short</code>.
     * @see    #parseShort(CharSequence, int)
     */
    public static short parseShort(CharSequence csq) {
        return parseShort(csq, 10);
    }

    /**
     * Parses the specified character sequence as a signed <code>short</code> 
     * in the specified radix.
     *
     * @param  csq the character sequence to parse.
     * @param  radix the radix to be used while parsing.
     * @return the corresponding <code>short</code>.
     * @throws NumberFormatException if the specified character sequence
     *         does not contain a parsable <code>short</code>.
     */
    public static short parseShort(CharSequence csq, int radix) {
        int i = parseInt(csq, radix);
        if ((i < Short.MIN_VALUE) || (i > Short.MAX_VALUE))
            throw new NumberFormatException("Overflow");
        return (short) i;
    }

    /**
     * Parses the specified character sequence from the specified position 
     * as a signed <code>short</code> in the specified radix.
     *
     * @param  csq the character sequence to parse.
     * @param  radix the radix to be used while parsing.
     * @param cursor the current cursor position (being maintained).
     * @return the corresponding <code>short</code>.
     * @throws NumberFormatException if the specified character sequence
     *         does not contain a parsable <code>short</code>.
     */
    public static short parseShort(CharSequence csq, int radix, Cursor cursor) {
        int i = parseInt(csq, radix, cursor);
        if ((i < Short.MIN_VALUE) || (i > Short.MAX_VALUE))
            throw new NumberFormatException("Overflow");
        return (short) i;
    }

    /**
     * Parses the specified character sequence as a signed <code>int</code>.
     *
     * @param  csq the character sequence to parse.
     * @return <code>parseInt(csq, 10)</code>
     * @throws NumberFormatException if the specified character sequence
     *         does not contain a parsable <code>int</code>.
     * @see    #parseInt(CharSequence, int)
     */
    public static int parseInt(CharSequence csq) {
        return parseInt(csq, 10);
    }

    /**
     * Parses the specified character sequence as a signed <code>int</code> 
     * in the specified radix.
     *
     * @param  csq the character sequence to parse.
     * @param  radix the radix to be used while parsing.
     * @return the corresponding <code>int</code>.
     * @throws NumberFormatException if the specified character sequence
     *         does not contain a parsable <code>int</code>.
     */
    public static int parseInt(CharSequence csq, int radix) {
        return parseInt(csq, radix, null);
    }

    /**
     * Parses the specified character sequence from the specified position
     * as a signed <code>int</code> in the specified radix.
     *
     * @param  csq the character sequence to parse.
     * @param  radix the radix to be used while parsing.
     * @param  cursor the current cursor position (being maintained) or
     *         <code>null</code> if the whole character sequence is parsed.
     * @return the corresponding <code>int</code>.
     * @throws NumberFormatException if the specified character sequence
     *         does not contain a parsable <code>int</code>.
     */
    public static int parseInt(CharSequence csq, int radix, Cursor cursor) {
        // Avoids dynamic cost of CharSequence.charAt
        if (csq instanceof CharArray)
            return parseIntCharArray((CharArray) csq, radix, cursor);
        if (csq instanceof TextBuilder)
            return parseIntTextBuilder((TextBuilder) csq, radix, cursor);
        if (csq instanceof Text)
            return parseIntText((Text) csq, radix, cursor);
        if (((Object) csq) instanceof String)
            return parseIntString((String) ((Object) csq), radix, cursor);
        return parseIntCharSequence(csq, radix, cursor);
    }

    private static int parseIntCharArray(CharArray csq, int radix, Cursor cursor) {
        // Parsing block identical for all CharSequences.
        final int start = (cursor != null) ? cursor.getIndex() : 0;
        final int length = csq.length();
        boolean isNegative = false;
        int result = 0; // Accumulates negatively (avoid MIN_VALUE overflow).
        int i = start;
        for (; i < length; i++) {
            char c = csq.charAt(i);
            int digit = (c <= '9') ? c - '0'
                    : ((c <= 'Z') && (c >= 'A')) ? c - 'A' + 10
                            : ((c <= 'z') && (c >= 'a')) ? c - 'a' + 10 : -1;
            if ((digit >= 0) && (digit < radix)) {
                int newResult = result * radix - digit;
                if (newResult > result)
                    throw new NumberFormatException("Overflow");
                result = newResult;
            } else if ((c == '-') && (i == start)) {
                isNegative = true;
            } else if ((c == '+') && (i == start)) {
                // Ok.
            } else {
                if (cursor == null)
                    throw new NumberFormatException("Incomplete parsing");
                break; // Done.
            }
        }
        // Requires one valid digit character and checks for opposite overflow.
        if ((result == 0) && ((length == 0) || (csq.charAt(i - 1) != '0')))
            throw new NumberFormatException("Cannot parse " + csq + " as int");
        if ((result == Integer.MIN_VALUE) && !isNegative)
            throw new NumberFormatException("Overflow");
        if (cursor != null)
            cursor.setIndex(i);
        return isNegative ? result : -result;
    }

    private static int parseIntTextBuilder(TextBuilder csq, int radix,
            Cursor cursor) {
        // Parsing block identical for all CharSequences.
        final int start = (cursor != null) ? cursor.getIndex() : 0;
        final int length = csq.length();
        boolean isNegative = false;
        int result = 0; // Accumulates negatively (avoid MIN_VALUE overflow).
        int i = start;
        for (; i < length; i++) {
            char c = csq.charAt(i);
            int digit = (c <= '9') ? c - '0'
                    : ((c <= 'Z') && (c >= 'A')) ? c - 'A' + 10
                            : ((c <= 'z') && (c >= 'a')) ? c - 'a' + 10 : -1;
            if ((digit >= 0) && (digit < radix)) {
                int newResult = result * radix - digit;
                if (newResult > result)
                    throw new NumberFormatException("Overflow");
                result = newResult;
            } else if ((c == '-') && (i == start)) {
                isNegative = true;
            } else if ((c == '+') && (i == start)) {
                // Ok.
            } else {
                if (cursor == null)
                    throw new NumberFormatException("Incomplete parsing");
                break; // Done.
            }
        }
        // Requires one valid digit character and checks for opposite overflow.
        if ((result == 0) && ((length == 0) || (csq.charAt(i - 1) != '0')))
            throw new NumberFormatException("Cannot parse " + csq + " as int");
        if ((result == Integer.MIN_VALUE) && !isNegative)
            throw new NumberFormatException("Overflow");
        if (cursor != null)
            cursor.setIndex(i);
        return isNegative ? result : -result;
    }

    private static int parseIntText(Text csq, int radix, Cursor cursor) {
        // Parsing block identical for all CharSequences.
        final int start = (cursor != null) ? cursor.getIndex() : 0;
        final int length = csq.length();
        boolean isNegative = false;
        int result = 0; // Accumulates negatively (avoid MIN_VALUE overflow).
        int i = start;
        for (; i < length; i++) {
            char c = csq.charAt(i);
            int digit = (c <= '9') ? c - '0'
                    : ((c <= 'Z') && (c >= 'A')) ? c - 'A' + 10
                            : ((c <= 'z') && (c >= 'a')) ? c - 'a' + 10 : -1;
            if ((digit >= 0) && (digit < radix)) {
                int newResult = result * radix - digit;
                if (newResult > result)
                    throw new NumberFormatException("Overflow");
                result = newResult;
            } else if ((c == '-') && (i == start)) {
                isNegative = true;
            } else if ((c == '+') && (i == start)) {
                // Ok.
            } else {
                if (cursor == null)
                    throw new NumberFormatException("Incomplete parsing");
                break; // Done.
            }
        }
        // Requires one valid digit character and checks for opposite overflow.
        if ((result == 0) && ((length == 0) || (csq.charAt(i - 1) != '0')))
            throw new NumberFormatException("Cannot parse " + csq + " as int");
        if ((result == Integer.MIN_VALUE) && !isNegative)
            throw new NumberFormatException("Overflow");
        if (cursor != null)
            cursor.setIndex(i);
        return isNegative ? result : -result;
    }

    private static int parseIntString(String csq, int radix, Cursor cursor) {
        // Parsing block identical for all CharSequences.
        final int start = (cursor != null) ? cursor.getIndex() : 0;
        final int length = csq.length();
        boolean isNegative = false;
        int result = 0; // Accumulates negatively (avoid MIN_VALUE overflow).
        int i = start;
        for (; i < length; i++) {
            char c = csq.charAt(i);
            int digit = (c <= '9') ? c - '0'
                    : ((c <= 'Z') && (c >= 'A')) ? c - 'A' + 10
                            : ((c <= 'z') && (c >= 'a')) ? c - 'a' + 10 : -1;
            if ((digit >= 0) && (digit < radix)) {
                int newResult = result * radix - digit;
                if (newResult > result)
                    throw new NumberFormatException("Overflow");
                result = newResult;
            } else if ((c == '-') && (i == start)) {
                isNegative = true;
            } else if ((c == '+') && (i == start)) {
                // Ok.
            } else {
                if (cursor == null)
                    throw new NumberFormatException("Incomplete parsing");
                break; // Done.
            }
        }
        // Requires one valid digit character and checks for opposite overflow.
        if ((result == 0) && ((length == 0) || (csq.charAt(i - 1) != '0')))
            throw new NumberFormatException("Cannot parse " + csq + " as int");
        if ((result == Integer.MIN_VALUE) && !isNegative)
            throw new NumberFormatException("Overflow");
        if (cursor != null)
            cursor.setIndex(i);
        return isNegative ? result : -result;
    }

    private static int parseIntCharSequence(CharSequence csq, int radix,
            Cursor cursor) {
        // Parsing block identical for all CharSequences.
        final int start = (cursor != null) ? cursor.getIndex() : 0;
        final int length = csq.length();
        boolean isNegative = false;
        int result = 0; // Accumulates negatively (avoid MIN_VALUE overflow).
        int i = start;
        for (; i < length; i++) {
            char c = csq.charAt(i);
            int digit = (c <= '9') ? c - '0'
                    : ((c <= 'Z') && (c >= 'A')) ? c - 'A' + 10
                            : ((c <= 'z') && (c >= 'a')) ? c - 'a' + 10 : -1;
            if ((digit >= 0) && (digit < radix)) {
                int newResult = result * radix - digit;
                if (newResult > result)
                    throw new NumberFormatException("Overflow");
                result = newResult;
            } else if ((c == '-') && (i == start)) {
                isNegative = true;
            } else if ((c == '+') && (i == start)) {
                // Ok.
            } else {
                if (cursor == null)
                    throw new NumberFormatException("Incomplete parsing");
                break; // Done.
            }
        }
        // Requires one valid digit character and checks for opposite overflow.
        if ((result == 0) && ((length == 0) || (csq.charAt(i - 1) != '0')))
            throw new NumberFormatException("Cannot parse " + csq + " as int");
        if ((result == Integer.MIN_VALUE) && !isNegative)
            throw new NumberFormatException("Overflow");
        if (cursor != null)
            cursor.setIndex(i);
        return isNegative ? result : -result;
    }

    /**
     * Parses the specified character sequence as a decimal <code>long</code>.
     *
     * @param  csq the character sequence to parse.
     * @return <code>parseLong(csq, 10)</code>
     * @throws NumberFormatException if the specified character sequence
     *         does not contain a parsable <code>long</code>.
     * @see    #parseLong(CharSequence, int)
     */
    public static long parseLong(CharSequence csq) {
        return parseLong(csq, 10);
    }

    /**
     * Parses the specified character sequence as a signed <code>long</code>
     * in the specified radix.
     *
     * @param  csq the character sequence to parse.
     * @param  radix the radix to be used while parsing.
     * @return the corresponding <code>long</code>.
     * @throws NumberFormatException if the specified character sequence
     *         does not contain a parsable <code>long</code>.
     */
    public static long parseLong(CharSequence csq, int radix) {
        return parseLong(csq, radix, null);
    }

    /**
     * Parses the specified character sequence from the specified position 
     * as a signed <code>long</code> in the specified radix.
     *
     * @param  csq the character sequence to parse.
     * @param  radix the radix to be used while parsing.
     * @param  cursor the current cursor position (being maintained) or
     *         <code>null</code> if the whole character sequence is parsed.
     * @return the corresponding <code>long</code>.
     * @throws NumberFormatException if the specified character sequence
     *         does not contain a parsable <code>long</code>.
     */
    public static long parseLong(CharSequence csq, int radix, Cursor cursor) {
        // Avoids dynamic cost of CharSequence.charAt
        if (csq instanceof CharArray)
            return parseLongCharArray((CharArray) csq, radix, cursor);
        if (csq instanceof TextBuilder)
            return parseLongTextBuilder((TextBuilder) csq, radix, cursor);
        if (csq instanceof Text)
            return parseLongText((Text) csq, radix, cursor);
        if (((Object) csq) instanceof String)
            return parseLongString((String) ((Object) csq), radix, cursor);
        return parseLongCharSequence(csq, radix, cursor);
    }

    private static long parseLongCharArray(CharArray csq, int radix,
            Cursor cursor) {
        // Parsing block identical for all CharSequences.
        final int start = (cursor != null) ? cursor.getIndex() : 0;
        final int length = csq.length();
        boolean isNegative = false;
        long result = 0; // Accumulates negatively (avoid MIN_VALUE overflow).
        int i = start;
        for (; i < length; i++) {
            char c = csq.charAt(i);
            int digit = (c <= '9') ? c - '0'
                    : ((c <= 'Z') && (c >= 'A')) ? c - 'A' + 10
                            : ((c <= 'z') && (c >= 'a')) ? c - 'a' + 10 : -1;
            if ((digit >= 0) && (digit < radix)) {
                long newResult = result * radix - digit;
                if (newResult > result)
                    throw new NumberFormatException("Overflow");
                result = newResult;
            } else if ((c == '-') && (i == start)) {
                isNegative = true;
            } else if ((c == '+') && (i == start)) {
                // Ok.
            } else {
                if (cursor == null)
                    throw new NumberFormatException("Incomplete parsing");
                break; // Done.
            }
        }
        // Requires one valid digit character and checks for opposite overflow.
        if ((result == 0) && ((length == 0) || (csq.charAt(i - 1) != '0')))
            throw new NumberFormatException("Cannot parse " + csq + " as int");
        if ((result == Long.MIN_VALUE) && !isNegative)
            throw new NumberFormatException("Overflow");
        if (cursor != null)
            cursor.setIndex(i);
        return isNegative ? result : -result;
    }

    private static long parseLongTextBuilder(TextBuilder csq, int radix,
            Cursor cursor) {
        // Parsing block identical for all CharSequences.
        final int start = (cursor != null) ? cursor.getIndex() : 0;
        final int length = csq.length();
        boolean isNegative = false;
        long result = 0; // Accumulates negatively (avoid MIN_VALUE overflow).
        int i = start;
        for (; i < length; i++) {
            char c = csq.charAt(i);
            int digit = (c <= '9') ? c - '0'
                    : ((c <= 'Z') && (c >= 'A')) ? c - 'A' + 10
                            : ((c <= 'z') && (c >= 'a')) ? c - 'a' + 10 : -1;
            if ((digit >= 0) && (digit < radix)) {
                long newResult = result * radix - digit;
                if (newResult > result)
                    throw new NumberFormatException("Overflow");
                result = newResult;
            } else if ((c == '-') && (i == start)) {
                isNegative = true;
            } else if ((c == '+') && (i == start)) {
                // Ok.
            } else {
                if (cursor == null)
                    throw new NumberFormatException("Incomplete parsing");
                break; // Done.
            }
        }
        // Requires one valid digit character and checks for opposite overflow.
        if ((result == 0) && ((length == 0) || (csq.charAt(i - 1) != '0')))
            throw new NumberFormatException("Cannot parse " + csq + " as int");
        if ((result == Long.MIN_VALUE) && !isNegative)
            throw new NumberFormatException("Overflow");
        if (cursor != null)
            cursor.setIndex(i);
        return isNegative ? result : -result;
    }

    private static long parseLongText(Text csq, int radix, Cursor cursor) {
        // Parsing block identical for all CharSequences.
        final int start = (cursor != null) ? cursor.getIndex() : 0;
        final int length = csq.length();
        boolean isNegative = false;
        long result = 0; // Accumulates negatively (avoid MIN_VALUE overflow).
        int i = start;
        for (; i < length; i++) {
            char c = csq.charAt(i);
            int digit = (c <= '9') ? c - '0'
                    : ((c <= 'Z') && (c >= 'A')) ? c - 'A' + 10
                            : ((c <= 'z') && (c >= 'a')) ? c - 'a' + 10 : -1;
            if ((digit >= 0) && (digit < radix)) {
                long newResult = result * radix - digit;
                if (newResult > result)
                    throw new NumberFormatException("Overflow");
                result = newResult;
            } else if ((c == '-') && (i == start)) {
                isNegative = true;
            } else if ((c == '+') && (i == start)) {
                // Ok.
            } else {
                if (cursor == null)
                    throw new NumberFormatException("Incomplete parsing");
                break; // Done.
            }
        }
        // Requires one valid digit character and checks for opposite overflow.
        if ((result == 0) && ((length == 0) || (csq.charAt(i - 1) != '0')))
            throw new NumberFormatException("Cannot parse " + csq + " as int");
        if ((result == Long.MIN_VALUE) && !isNegative)
            throw new NumberFormatException("Overflow");
        if (cursor != null)
            cursor.setIndex(i);
        return isNegative ? result : -result;
    }

    private static long parseLongString(String csq, int radix, Cursor cursor) {
        // Parsing block identical for all CharSequences.
        final int start = (cursor != null) ? cursor.getIndex() : 0;
        final int length = csq.length();
        boolean isNegative = false;
        long result = 0; // Accumulates negatively (avoid MIN_VALUE overflow).
        int i = start;
        for (; i < length; i++) {
            char c = csq.charAt(i);
            int digit = (c <= '9') ? c - '0'
                    : ((c <= 'Z') && (c >= 'A')) ? c - 'A' + 10
                            : ((c <= 'z') && (c >= 'a')) ? c - 'a' + 10 : -1;
            if ((digit >= 0) && (digit < radix)) {
                long newResult = result * radix - digit;
                if (newResult > result)
                    throw new NumberFormatException("Overflow");
                result = newResult;
            } else if ((c == '-') && (i == start)) {
                isNegative = true;
            } else if ((c == '+') && (i == start)) {
                // Ok.
            } else {
                if (cursor == null)
                    throw new NumberFormatException("Incomplete parsing");
                break; // Done.
            }
        }
        // Requires one valid digit character and checks for opposite overflow.
        if ((result == 0) && ((length == 0) || (csq.charAt(i - 1) != '0')))
            throw new NumberFormatException("Cannot parse " + csq + " as int");
        if ((result == Long.MIN_VALUE) && !isNegative)
            throw new NumberFormatException("Overflow");
        if (cursor != null)
            cursor.setIndex(i);
        return isNegative ? result : -result;
    }

    private static long parseLongCharSequence(CharSequence csq, int radix,
            Cursor cursor) {
        // Parsing block identical for all CharSequences.
        final int start = (cursor != null) ? cursor.getIndex() : 0;
        final int length = csq.length();
        boolean isNegative = false;
        long result = 0; // Accumulates negatively (avoid MIN_VALUE overflow).
        int i = start;
        for (; i < length; i++) {
            char c = csq.charAt(i);
            int digit = (c <= '9') ? c - '0'
                    : ((c <= 'Z') && (c >= 'A')) ? c - 'A' + 10
                            : ((c <= 'z') && (c >= 'a')) ? c - 'a' + 10 : -1;
            if ((digit >= 0) && (digit < radix)) {
                long newResult = result * radix - digit;
                if (newResult > result)
                    throw new NumberFormatException("Overflow");
                result = newResult;
            } else if ((c == '-') && (i == start)) {
                isNegative = true;
            } else if ((c == '+') && (i == start)) {
                // Ok.
            } else {
                if (cursor == null)
                    throw new NumberFormatException("Incomplete parsing");
                break; // Done.
            }
        }
        // Requires one valid digit character and checks for opposite overflow.
        if ((result == 0) && ((length == 0) || (csq.charAt(i - 1) != '0')))
            throw new NumberFormatException("Cannot parse " + csq + " as int");
        if ((result == Long.MIN_VALUE) && !isNegative)
            throw new NumberFormatException("Overflow");
        if (cursor != null)
            cursor.setIndex(i);
        return isNegative ? result : -result;
    }

    /**
     * Parses the specified character sequence as a <code>float</code>.
     *
     * @param  csq the character sequence to parse.
     * @return the float number represented by the specified character sequence.
     /*@JVM-1.1+@
     public static float parseFloat(CharSequence csq) {
     return (float) parseDouble(csq);
     }
     /**/

    /**
     * Parses the specified character sequence from the specified position 
     * as a <code>float</code>.
     *
     * @param  csq the character sequence to parse.
     * @param  cursor the current cursor position (being maintained).
     * @return the float number represented by the specified character sequence.
     /*@JVM-1.1+@
     public static float parseFloat(CharSequence csq, Cursor cursor) {
     return (float) parseDouble(csq, cursor);
     }
     /**/

    /**
     * Parses the specified character sequence as a <code>double</code>.
     *
     * @param  csq the character sequence to parse.
     * @return the double number represented by this character sequence.
     * @throws NumberFormatException if the character sequence does not contain
     *         a parsable <code>double</code>.
     /*@JVM-1.1+@
     public static double parseDouble(CharSequence csq)
     throws NumberFormatException {
     Cursor cursor = (Cursor) CURSOR.get();
     cursor.setIndex(0);
     double value = parseDouble(csq, cursor);
     if (cursor.getIndex() != csq.length()) 
     throw new NumberFormatException("Incomplete parsing");
     return value;
     }
     /**/

    /**
     * Parses the specified character sequence from the specified position 
     * as a <code>double</code>.
     *
     * @param  csq the character sequence to parse.
     * @param  cursor the current cursor position (being maintained).
     * @return the double number represented by this character sequence.
     * @throws NumberFormatException if the character sequence does not contain
     *         a parsable <code>double</code>.
     /*@JVM-1.1+@
     public static double parseDouble(CharSequence csq, Cursor cursor)
     throws NumberFormatException {
     try {
     int i = cursor.getIndex();
     // Reads the sign now (which might not be the sign of the decimal part, e.g. -0.XXX).
     boolean isNegative = (csq.charAt(i) == '-') ? true : false;
     i += (isNegative || (csq.charAt(i) == '+')) ? 1 : 0;
     char c0 = csq.charAt(i);
     if ((c0 == 'N') && startWith(csq, i, "NaN")) {
     cursor.setIndex(i + 3);
     return Double.NaN;
     }
     if ((c0 == 'I') && startWith(csq, i, "Infinity")) {
     cursor.setIndex(i + 8);
     return isNegative ? Double.NEGATIVE_INFINITY : Double.POSITIVE_INFINITY;
     }
     cursor.setIndex(i);
     long decimal = parseLong(csq, 10, cursor);
     final int length = csq.length();

     // Reads fraction.
     long fraction = 0;
     int fractionLength = 0;
     if (cursor.getIndex() < length) {
     char c = csq.charAt(cursor.getIndex());
     if (c == '.') {
     cursor.increment();
     int startFraction = cursor.getIndex();
     fraction = TypeFormat.parseLong(csq, 10, cursor);
     fractionLength = cursor.getIndex() - startFraction;
     }
     }

     // Reads exponent.
     int exponent = 0;
     if (cursor.getIndex() < length) {
     char c = csq.charAt(cursor.getIndex());
     if ((c == 'e') || (c == 'E')) {
     cursor.increment();
     exponent = TypeFormat.parseInt(csq, 10, cursor);
     }
     }

     // Adds decimal and fraction part.  
     double decimalAsDouble = javolution.lang.MathLib.toDouble(decimal, exponent);
     double fractionAsDouble = javolution.lang.MathLib.toDouble(
     decimal < 0 ? -fraction : fraction, exponent - fractionLength);
     
     return isNegative ? - decimalAsDouble - fractionAsDouble :
     decimalAsDouble + fractionAsDouble;
     
     } catch (IndexOutOfBoundsException e) {
     throw new NumberFormatException();
     }
     }
     /**/

    static boolean startWith(CharSequence csq, int start, String str) {
        for (int i = 0; i < str.length(); i++) {
            if ((start + i >= csq.length())
                    || csq.charAt(start + i) != str.charAt(i))
                return false;
        }
        return true;
    }

    /**
     * Formats the specified <code>boolean</code> and appends the resulting
     * text to the <code>Appendable</code> argument.
     *
     * @param  b a <code>boolean</code>.
     * @param  a the <code>Appendable</code> to append.
     * @return the specified <code>StringBuffer</code> object.
     * @throws IOException if an I/O exception occurs.
     * @see    #parseBoolean
     */
    public static Appendable format(boolean b, Appendable a) throws IOException {
        return b ? append(a, "true") : append(a, "false");
    }

    /**
     * Formats the specified <code>int</code> and appends the resulting
     * text (decimal representation) to the <code>Appendable</code> argument.
     *
     * <p> Note: This method is preferred to <code>Appendable.append(int)
     *           </code> as it does not create temporary <code>String</code>
     *           objects (several times faster for small numbers).</p>
     *
     * @param  i the <code>int</code> number.
     * @param  a the <code>Appendable</code> to append.
     * @return the specified <code>Appendable</code> object.
     * @throws IOException if an I/O exception occurs.
     * @see    #parseInt
     */
    public static Appendable format(int i, Appendable a) throws IOException {
        if (a instanceof TextBuilder)
            return format(i, (TextBuilder) a);
        if (i < 0) {
            if (i == Integer.MIN_VALUE) // Negation would overflow.
                return append(a, "-2147483648"); // 11 char max.
            i = -i;
            a.append('-');
        }
        if (i >= 100000) {
            int high = i / 100000;
            writeDigits(high, a, false);
            i -= high * 100000;
            writeDigits(i, a, true);
        } else {
            writeDigits(i, a, false);
        }
        return a;
    }

    static TextBuilder format(int i, TextBuilder a) {
        if (i < 0) {
            if (i == Integer.MIN_VALUE) // Negation would overflow.
                return a.append("-2147483648"); // 11 char max.
            i = -i;
            a.append('-');
        }
        if (i >= 100000) {
            int high = i / 100000;
            writeDigits(high, a, false);
            i -= high * 100000;
            writeDigits(i, a, true);
        } else {
            writeDigits(i, a, false);
        }
        return a;
    }

    private static final int[] INT_POW_10 = new int[10];
    static {
        int pow = 1;
        for (int i = 0; i < 10; i++) {
            INT_POW_10[i] = pow;
            pow *= 10;
        }
    }

    /**
     * Formats the specified <code>int</code> in the specified radix and appends
     * the resulting text to the <code>Appendable</code> argument.
     *
     * @param  i the <code>int</code> number.
     * @param  radix the radix.
     * @param  a the <code>Appendable</code> to append.
     * @return the specified <code>Appendable</code> object.
     * @throws  IllegalArgumentException if radix is not in [2 .. 36] range.
     * @throws IOException if an I/O exception occurs.
     * @see    #parseInt(CharSequence, int)
     */
    public static Appendable format(int i, int radix, Appendable a)
            throws IOException {
        if (radix == 10) {
            return format(i, a); // Faster version.
        } else if (radix < 2 || radix > 36) {
            throw new IllegalArgumentException("radix: " + radix);
        }
        if (i < 0) {
            a.append('-');
        } else {
            i = -i;
        }
        format2(i, radix, a);
        return a;
    }

    private static void format2(int i, int radix, Appendable a)
            throws IOException {
        if (i <= -radix) {
            format2(i / radix, radix, a);
            a.append(DIGIT_TO_CHAR[-(i % radix)]);
        } else {
            a.append(DIGIT_TO_CHAR[-i]);
        }
    }

    /**
     * Formats the specified <code>long</code> and appends the resulting
     * text (decimal representation) to the <code>Appendable</code> argument.
     *
     * <p> Note: This method is preferred to <code>Appendable.append(long)
     *           </code> as it does not create temporary <code>String</code>
     *           objects (several times faster for small numbers).</p>
     *
     * @param  l the <code>long</code> number.
     * @param  a the <code>Appendable</code> to append.
     * @return the specified <code>Appendable</code> object.
     * @throws IOException if an I/O exception occurs.
     * @see    #parseLong
     */
    public static Appendable format(long l, Appendable a) throws IOException {
        if (a instanceof TextBuilder)
            return format(l, (TextBuilder) a);
        if (l < 0) {
            if (l == Long.MIN_VALUE) // Negation would overflow.
                return append(a, "-9223372036854775808"); // 20 characters max.
            l = -l;
            a.append('-');
        }
        boolean writeZero = false;
        if (l >= 1000000000000000L) {
            int high = (int) (l / 1000000000000000L);
            writeDigits(high, a, writeZero);
            l -= high * 1000000000000000L;
            writeZero = true;
        }
        if (writeZero || (l >= 10000000000L)) {
            int high = (int) (l / 10000000000L);
            writeDigits(high, a, writeZero);
            l -= high * 10000000000L;
            writeZero = true;
        }
        if (writeZero || (l >= 100000)) {
            int high = (int) (l / 100000);
            writeDigits(high, a, writeZero);
            l -= high * 100000L;
            writeZero = true;
        }
        writeDigits((int) l, a, writeZero);
        return a;
    }

    static TextBuilder format(long l, TextBuilder a) {
        if (l < 0) {
            if (l == Long.MIN_VALUE) // Negation would overflow.
                return a.append("-9223372036854775808"); // 20 characters max.
            l = -l;
            a.append('-');
        }
        boolean writeZero = false;
        if (l >= 1000000000000000L) {
            int high = (int) (l / 1000000000000000L);
            writeDigits(high, a, writeZero);
            l -= high * 1000000000000000L;
            writeZero = true;
        }
        if (writeZero || (l >= 10000000000L)) {
            int high = (int) (l / 10000000000L);
            writeDigits(high, a, writeZero);
            l -= high * 10000000000L;
            writeZero = true;
        }
        if (writeZero || (l >= 100000)) {
            int high = (int) (l / 100000);
            writeDigits(high, a, writeZero);
            l -= high * 100000L;
            writeZero = true;
        }
        writeDigits((int) l, a, writeZero);
        return a;
    }

    private static final long[] LONG_POW_10 = new long[19];
    static {
        long pow = 1;
        for (int i = 0; i < 19; i++) {
            LONG_POW_10[i] = pow;
            pow *= 10;
        }
    }

    /**
     * Formats the specified <code>long</code> in the specified radix and
     * appends the resulting text to the <code>Appendable</code> argument.
     *
     * @param  l the <code>long</code> number.
     * @param  radix the radix.
     * @param  a the <code>Appendable</code> to append.
     * @return the specified <code>Appendable</code> object.
     * @throws  IllegalArgumentException if radix is not in [2 .. 36] range.
     * @throws IOException if an I/O exception occurs.
     * @see    #parseLong(CharSequence, int)
     */
    public static Appendable format(long l, int radix, Appendable a)
            throws IOException {
        if (radix == 10) {
            return format(l, a); // Faster version.
        } else if (radix < 2 || radix > 36) {
            throw new IllegalArgumentException("radix: " + radix);
        }
        if (l < 0) {
            a.append('-');
        } else {
            l = -l;
        }
        format2(l, radix, a);
        return a;
    }

    private static void format2(long l, int radix, Appendable a)
            throws IOException {
        if (l <= -radix) {
            format2(l / radix, radix, a);
            a.append(DIGIT_TO_CHAR[(int) -(l % radix)]);
        } else {
            a.append(DIGIT_TO_CHAR[(int) -l]);
        }
    }

    /**
     * Formats the specified <code>float</code> value.
     *
     * @param  value the <code>float</code> value.
     * @param  a the <code>Appendable</code> to append.
     * @return <code>format(value, 10, Math.abs(value) > 1E7, false, sb)</code>
     * @throws IOException if an I/O exception occurs.
     * @see    #format(double, int , boolean, boolean, Appendable)
     /*@JVM-1.1+@
     public static Appendable format(float value, Appendable a)
     throws IOException {
     return format(value, 10, javolution.lang.MathLib.abs(value) > 1E7, false, a);
     }
     /**/

    /**
     * Formats the specified <code>double</code> value.
     *
     * <p> Note : This method is preferred to <code>Double.toString(double)
     *            </code> or even <code>String.valueOf(double)</code> as it
     *            does not create temporary <code>String</code> or <code>
     *            FloatingDecimal</code> objects (several times faster,
     *            e.g. 15x faster for <code>Double.MAX_VALUE</code>).</p>
     *
     * @param  value the <code>double</code> value.
     * @param  a the <code>Appendable</code> to append.
     * @return <code>format(value, 17, Math.abs(value) > 1E7, false, sb)</code>
     * @throws IOException if an I/O exception occurs.
     * @see    #format(double, int , boolean, boolean, Appendable)
     /*@JVM-1.1+@
     public static Appendable format(double value, Appendable a)
     throws IOException {
     return format(value, 17, javolution.lang.MathLib.abs(value) > 1E7, false, a);
     }
     /**/

    /*  Exact version (but too slow) 
     public static Appendable format(double value, int digits,
     boolean scientific, boolean showZero, Appendable a) throws IOException {
     if ((digits > 19) || (digits <= 0))
     throw new IllegalArgumentException("digits: " + digits);
     if (value != value) { // NaN
     return append(a, "NaN");
     } else if (value == POSITIVE_INFINITY) {
     return append(a, "Infinity");
     } else if (value == NEGATIVE_INFINITY) {
     return append(a, "-Infinity");
     } else if (value == 0.0) {
     if (digits == 1)
     return append(a, "0");
     if (!showZero)
     return append(a, "0.0");
     append(a, "0.0");
     for (int i = 2; i < digits; i++) {
     a.append('0');
     }
     return a;
     }
     if (value < 0) {
     value = -value;
     a.append('-');
     }
     long xBits = Double.doubleToLongBits(value);
     int n = ((int)(xBits>>52))-1075;
     long m = xBits & 0x000fffffffffffffL | 0x0010000000000000L;
     int e = -2;
     m *= 100;
     while (n != 0) {
     if (n < 0) {
     if ((m & 0xF800000000000000L) == 0L) {
     m *= 5;
     e--;
     n++;
     } else {
     m >>= 1;
     n++;
     }
     } else {
     if ((m & 0xC000000000000000L) == 0L) {
     m <<= 1;
     n--;
     } else {
     m /= 10;
     e++;
     }
     }
     }
     digits = MathLib.digitLength(m);
     e += digits;
     long mantissa = m; 
     
     if (scientific || (e <= 0) || (e > digits)) {
     // Scientific notation has to be used ("x.xxxEyy").
     format(mantissa / LONG_POW_10[digits - 1], a);
     formatFraction(mantissa % LONG_POW_10[digits - 1], digits - 1,
     showZero, a);
     a.append('E');
     format(e - 1, a);
     } else if (e == digits) { // Dot at last position ("xxxxx").
     format(mantissa, a);
     } else { // Dot within the string ("xxxx.xxxxx").
     format(mantissa / LONG_POW_10[digits - e], a);
     formatFraction(mantissa % LONG_POW_10[digits - e], digits
     - e, showZero, a);
     }
     return a;
     }
     */

    /**
     * Formats the specified <code>double</code> value according to the 
     * specified formatting arguments.
     *
     * @param  value the <code>double</code> value.
     * @param  digits the number of significative digits (excludes exponent).
     * @param  scientific <code>true</code> to forces the use of the scientific 
     *         notation (e.g. <code>1.23E3</code>); <code>false</code> 
     *         otherwise. 
     * @param  showZero <code>true</code> if trailing fractional zeros are 
     *         represented; <code>false</code> otherwise.
     * @param  a the <code>Appendable</code> to append.
     * @return the specified <code>Appendable</code> object.
     * @throws IllegalArgumentException if <code>((digits > 19) || 
     *         (digits <= 0))</code>)
     * @throws IOException if an I/O exception occurs.
     /*@JVM-1.1+@
     public static Appendable format(double value, int digits, 
     boolean scientific, boolean showZero, Appendable a) throws IOException {
     if ((digits > 19) || (digits <= 0))
     throw new IllegalArgumentException("digits: " + digits);
     if (value != value) { // NaN
     return append(a, "NaN");
     } else if (value == POSITIVE_INFINITY) {
     return append(a, "Infinity");
     } else if (value == NEGATIVE_INFINITY) {
     return append(a, "-Infinity");
     } else if (value == 0.0) {
     if (digits == 1)
     return append(a, "0");
     if (!showZero)
     return append(a, "0.0");
     append(a, "0.0");
     for (int i = 2; i < digits; i++) {
     a.append('0');
     }
     return a;
     }
     if (value < 0) {
     value = -value;
     a.append('-');
     }
     // Find the exponent e such as: value == 0.xxx * 10^e
     int e = (value >= 1.0) ? 1 + minPow10(value) : - minPow10(1.0 / value);
     double digitValue = value * javolution.lang.MathLib.toDouble(1, digits - e);
     long mantissa = (long) (digitValue + 0.5);
     if (scientific || (e <= 0) || (e > digits)) {
     // Scientific notation has to be used ("x.xxxEyy").
     format(mantissa / LONG_POW_10[digits - 1], a);
     formatFraction(mantissa % LONG_POW_10[digits - 1], digits - 1,
     showZero, a);
     a.append('E');
     format(e - 1, a);
     } else if (e == digits) { // Dot at last position ("xxxxx").
     format(mantissa, a);
     } else { // Dot within the string ("xxxx.xxxxx").
     format(mantissa / LONG_POW_10[digits - e], a);
     formatFraction(mantissa % LONG_POW_10[digits - e], digits
     - e, showZero, a);
     }
     return a;
     }


     private static final double POSITIVE_INFINITY = 1.0 / 0.0;

     private static final double NEGATIVE_INFINITY = -1.0 / 0.0;

     private static void formatFraction(long fraction, int digits,
     boolean showZero, Appendable a) throws IOException {
     if (digits == 0)
     return;
     a.append('.');
     for (int i = digits; i > 0;) {
     long pow10 = LONG_POW_10[--i];
     int digit = (int) (fraction / pow10);
     fraction -= digit * pow10;
     a.append(DIGIT_TO_CHAR[digit]);
     if ((fraction == 0) && !showZero) {
     return; // No more than one trailing zero.
     }
     }
     }
     // Returns e such as 10^e <= value < 10^(e+1), value >= 1.0
     private static int minPow10(double value) {
     int minE = 0;
     int maxE = 308;
     while (maxE - minE > 1) {
     final int exp = (minE + maxE) >> 1;
     if (value >= javolution.lang.MathLib.toDouble(1, exp)) {
     minE = exp;
     } else {
     maxE = exp;
     }
     }
     return minE;
     }
     /**/

    /**
     * Appends the specified string argument to the specified appendable
     * for backward compatibility when {@link String} was not a 
     * {@link CharSequence}.
     *
     * @param  to the appendable.
     * @param  str the string to append.
     * @return the specified appendable
     * @throws IOException if an I/O exception occurs.
     */
    private static Appendable append(Appendable to, String str)
            throws IOException {
        for (int i = 0; i < str.length(); i++) {
            to.append(str.charAt(i));
        }
        return to;
    }

    /**
     * Appends the specified positive integer [0 .. 99999] in decimal
     * representation to the specified text builder.
     *
     * @param i the integer to write.
     * @param a the text builder appendable.
     * @param writeZero <code>true</code> if leading zero are included;
     *       <code>false</code> otherwise.
     */
    private static void writeDigits(int i, TextBuilder a, boolean writeZero) {
        {
            final int e = 10000;
            if (i >= e) {
                writeZero = true;
                if (i >= e * 5) {
                    if (i >= e * 8) {
                        if (i >= e * 9) {
                            a.append('9');
                            i -= e * 9;
                        } else {
                            a.append('8');
                            i -= e * 8;
                        }
                    } else { // [5 .. 8[
                        if (i >= e * 7) {
                            a.append('7');
                            i -= e * 7;
                        } else if (i >= e * 6) {
                            a.append('6');
                            i -= e * 6;
                        } else {
                            a.append('5');
                            i -= e * 5;
                        }
                    }
                } else { // [1 ..5[
                    if (i >= e * 3) {
                        if (i >= e * 4) {
                            a.append('4');
                            i -= e * 4;
                        } else {
                            a.append('3');
                            i -= e * 3;
                        }
                    } else { // [1 .. 2[
                        if (i >= e * 2) {
                            a.append('2');
                            i -= e * 2;
                        } else {
                            a.append('1');
                            i -= e;
                        }
                    }
                }
            } else if (writeZero) {
                a.append('0');
            }
        }
        {
            final int e = 1000;
            if (i >= e) {
                writeZero = true;
                if (i >= e * 5) {
                    if (i >= e * 8) {
                        if (i >= e * 9) {
                            a.append('9');
                            i -= e * 9;
                        } else {
                            a.append('8');
                            i -= e * 8;
                        }
                    } else { // [5 .. 8[
                        if (i >= e * 7) {
                            a.append('7');
                            i -= e * 7;
                        } else if (i >= e * 6) {
                            a.append('6');
                            i -= e * 6;
                        } else {
                            a.append('5');
                            i -= e * 5;
                        }
                    }
                } else { // [1 ..5[
                    if (i >= e * 3) {
                        if (i >= e * 4) {
                            a.append('4');
                            i -= e * 4;
                        } else {
                            a.append('3');
                            i -= e * 3;
                        }
                    } else { // [1 .. 2[
                        if (i >= e * 2) {
                            a.append('2');
                            i -= e * 2;
                        } else {
                            a.append('1');
                            i -= e;
                        }
                    }
                }
            } else if (writeZero) {
                a.append('0');
            }
        }
        {
            final int e = 100;
            if (i >= e) {
                writeZero = true;
                if (i >= e * 5) {
                    if (i >= e * 8) {
                        if (i >= e * 9) {
                            a.append('9');
                            i -= e * 9;
                        } else {
                            a.append('8');
                            i -= e * 8;
                        }
                    } else { // [5 .. 8[
                        if (i >= e * 7) {
                            a.append('7');
                            i -= e * 7;
                        } else if (i >= e * 6) {
                            a.append('6');
                            i -= e * 6;
                        } else {
                            a.append('5');
                            i -= e * 5;
                        }
                    }
                } else { // [1 ..5[
                    if (i >= e * 3) {
                        if (i >= e * 4) {
                            a.append('4');
                            i -= e * 4;
                        } else {
                            a.append('3');
                            i -= e * 3;
                        }
                    } else { // [1 .. 2[
                        if (i >= e * 2) {
                            a.append('2');
                            i -= e * 2;
                        } else {
                            a.append('1');
                            i -= e;
                        }
                    }
                }
            } else if (writeZero) {
                a.append('0');
            }
        }
        {
            final int e = 10;
            if (i >= e) {
                writeZero = true;
                if (i >= e * 5) {
                    if (i >= e * 8) {
                        if (i >= e * 9) {
                            a.append('9');
                            i -= e * 9;
                        } else {
                            a.append('8');
                            i -= e * 8;
                        }
                    } else { // [5 .. 8[
                        if (i >= e * 7) {
                            a.append('7');
                            i -= e * 7;
                        } else if (i >= e * 6) {
                            a.append('6');
                            i -= e * 6;
                        } else {
                            a.append('5');
                            i -= e * 5;
                        }
                    }
                } else { // [1 ..5[
                    if (i >= e * 3) {
                        if (i >= e * 4) {
                            a.append('4');
                            i -= e * 4;
                        } else {
                            a.append('3');
                            i -= e * 3;
                        }
                    } else { // [1 .. 2[
                        if (i >= e * 2) {
                            a.append('2');
                            i -= e * 2;
                        } else {
                            a.append('1');
                            i -= e;
                        }
                    }
                }
            } else if (writeZero) {
                a.append('0');
            }
        }
        a.append(DIGIT_TO_CHAR[i]);
    }

    /**
     * Appends the specified positive integer [0 .. 99999] in decimal
     * representation to the specified appendable.
     *
     * @param i the integer to write.
     * @param a the appendable.
     * @param writeZero <code>true</code> if leading zero are included;
     *       <code>false</code> otherwise.
     * @throws IOException if an I/O exception occurs.
     */
    private static void writeDigits(int i, Appendable a, boolean writeZero)
            throws IOException {
        {
            final int e = 10000;
            if (i >= e) {
                writeZero = true;
                if (i >= e * 5) {
                    if (i >= e * 8) {
                        if (i >= e * 9) {
                            a.append('9');
                            i -= e * 9;
                        } else {
                            a.append('8');
                            i -= e * 8;
                        }
                    } else { // [5 .. 8[
                        if (i >= e * 7) {
                            a.append('7');
                            i -= e * 7;
                        } else if (i >= e * 6) {
                            a.append('6');
                            i -= e * 6;
                        } else {
                            a.append('5');
                            i -= e * 5;
                        }
                    }
                } else { // [1 ..5[
                    if (i >= e * 3) {
                        if (i >= e * 4) {
                            a.append('4');
                            i -= e * 4;
                        } else {
                            a.append('3');
                            i -= e * 3;
                        }
                    } else { // [1 .. 2[
                        if (i >= e * 2) {
                            a.append('2');
                            i -= e * 2;
                        } else {
                            a.append('1');
                            i -= e;
                        }
                    }
                }
            } else if (writeZero) {
                a.append('0');
            }
        }
        {
            final int e = 1000;
            if (i >= e) {
                writeZero = true;
                if (i >= e * 5) {
                    if (i >= e * 8) {
                        if (i >= e * 9) {
                            a.append('9');
                            i -= e * 9;
                        } else {
                            a.append('8');
                            i -= e * 8;
                        }
                    } else { // [5 .. 8[
                        if (i >= e * 7) {
                            a.append('7');
                            i -= e * 7;
                        } else if (i >= e * 6) {
                            a.append('6');
                            i -= e * 6;
                        } else {
                            a.append('5');
                            i -= e * 5;
                        }
                    }
                } else { // [1 ..5[
                    if (i >= e * 3) {
                        if (i >= e * 4) {
                            a.append('4');
                            i -= e * 4;
                        } else {
                            a.append('3');
                            i -= e * 3;
                        }
                    } else { // [1 .. 2[
                        if (i >= e * 2) {
                            a.append('2');
                            i -= e * 2;
                        } else {
                            a.append('1');
                            i -= e;
                        }
                    }
                }
            } else if (writeZero) {
                a.append('0');
            }
        }
        {
            final int e = 100;
            if (i >= e) {
                writeZero = true;
                if (i >= e * 5) {
                    if (i >= e * 8) {
                        if (i >= e * 9) {
                            a.append('9');
                            i -= e * 9;
                        } else {
                            a.append('8');
                            i -= e * 8;
                        }
                    } else { // [5 .. 8[
                        if (i >= e * 7) {
                            a.append('7');
                            i -= e * 7;
                        } else if (i >= e * 6) {
                            a.append('6');
                            i -= e * 6;
                        } else {
                            a.append('5');
                            i -= e * 5;
                        }
                    }
                } else { // [1 ..5[
                    if (i >= e * 3) {
                        if (i >= e * 4) {
                            a.append('4');
                            i -= e * 4;
                        } else {
                            a.append('3');
                            i -= e * 3;
                        }
                    } else { // [1 .. 2[
                        if (i >= e * 2) {
                            a.append('2');
                            i -= e * 2;
                        } else {
                            a.append('1');
                            i -= e;
                        }
                    }
                }
            } else if (writeZero) {
                a.append('0');
            }
        }
        {
            final int e = 10;
            if (i >= e) {
                writeZero = true;
                if (i >= e * 5) {
                    if (i >= e * 8) {
                        if (i >= e * 9) {
                            a.append('9');
                            i -= e * 9;
                        } else {
                            a.append('8');
                            i -= e * 8;
                        }
                    } else { // [5 .. 8[
                        if (i >= e * 7) {
                            a.append('7');
                            i -= e * 7;
                        } else if (i >= e * 6) {
                            a.append('6');
                            i -= e * 6;
                        } else {
                            a.append('5');
                            i -= e * 5;
                        }
                    }
                } else { // [1 ..5[
                    if (i >= e * 3) {
                        if (i >= e * 4) {
                            a.append('4');
                            i -= e * 4;
                        } else {
                            a.append('3');
                            i -= e * 3;
                        }
                    } else { // [1 .. 2[
                        if (i >= e * 2) {
                            a.append('2');
                            i -= e * 2;
                        } else {
                            a.append('1');
                            i -= e;
                        }
                    }
                }
            } else if (writeZero) {
                a.append('0');
            }
        }
        a.append(DIGIT_TO_CHAR[i]);
    }
}